/*
  Program: Pokrvn mky pomoc polyomoin
  Autor: Radek Pelnek  
  Ploha ke knize "Jak to vyeit",  www.radekpelanek.cz/?jtv

  Zadn projektu: Napite program pro een loh s polyominy. Program na
  vstupu nate dlky, kter je dovoleno pout, a clov obrazec, kter m bt
  z dlk vyskldn. Vstupem programu je vpis een (nap. v jednoduchm
  textovm znzornn), ppadn vpis vech een.

  Tento program spolu s piloenm souborem "tetronima" a "mrizkaX" e
  pokryt mek pomoc sady tetronim. Je vak psn relativn obecn a ml by jt
  upravit i pro jin problmy.

  Pouit programu: program oekv jako argument jmna dvou soubor (soubor s
  popisem dlkl a soubor s popisem tvaru mky), viz piloen soubory.
*/


#include <iostream>
#include <fstream>

#define MAXPSIZE 4    // maximln velikost jednoho dlku
#define MAXROT 8      // maximln poet rotac jednoho dlku
#define MAXPLAN 12    // maximln velikost plnu
#define MAXPIECES 5   // maximln poet rznch dlk

using namespace std;

// dlek vetn pevnho natoen
struct fixed_piece_t {
  int size;
  bool a[MAXPSIZE][MAXPSIZE];
};

// tda reprezentujc dlek vetn vech monch natoen
class free_piece_t {
  int size;
  bool a[MAXPSIZE][MAXPSIZE][MAXROT];
public:
  int positions;
  free_piece_t() { positions = 0; size = 0;}
  void read(ifstream&);
  void get_rotation(int, fixed_piece_t&);
};

// tda reprezentujc stav cel mky (sten pokryt dlky)
class state_t {
public:
  int n, m;
  int b[MAXPLAN][MAXPLAN];
  state_t() {n=m=0;}
  void read(ifstream&);
  void print();
  void copy(state_t&);
  bool match(int, int, fixed_piece_t&);
  void commit(int, int, fixed_piece_t&, int);
};

// naten dlku ze souboru
// neprovd dnou kontrolu korektnosti, pi patnm formtu se program nechov rozumn
void free_piece_t::read(ifstream& in) {
  char c;
  in >> size;
  in >> positions;
  for (int p = 0; p < positions; p++) {
    for (int j = 0; j < size; j++) 
      for (int i = 0; i < size; i++) {
	in >> c;
	a[i][j][p] = (c == '#');
      }
  }
}

// zskn konkrtnho natoen dlku
void free_piece_t::get_rotation(int p, fixed_piece_t& d) {
  d.size = size;
  for (int j = 0; j < size; j++) 
    for (int i = 0; i < size; i++) d.a[i][j] = a[i][j][p];
}

// naten tvaru celkov mky
// neprovd dnou kontrolu korektnosti, pi patnm formtu se program nechov rozumn
void state_t::read(ifstream& in) {
  char c;
  in >> n >> m;
  for (int j = 0; j < m; j++) 
    for (int i = 0; i < n; i++) {
      in >> c;
      b[i][j] = (c == '.') ? 0 : -1;
    }
}

// zkoprovn stavu cel mky
void state_t::copy(state_t& s) {
  n = s.n; m = s.m;
  for (int i=0; i<n; i++)
    for (int j=0; j<m;j++)
      b[i][j] = s.b[i][j];
}

// vpis mky
void state_t::print() {
  for (int j = 0; j < m; j++) {
    for (int i = 0; i < n; i++) {
      if (b[i][j] == -1) cout << '#';
      else cout << b[i][j];
    }
    cout << endl;
  }
  cout << "------------------"<<endl;  
}

// kontrola, zda lze dlek vloit do mky na zadanou pozici
bool state_t::match(int x, int y, fixed_piece_t& p) {
  bool ok = true;
  
  for (int i=0; i<p.size; i++)
    for (int j=0; j<p.size; j++)
      if (p.a[i][j] && (x+i>=n  || y+j>=m || b[x+i][y+j] != 0)) ok = false;
  return ok;
}

// vloen dlku do mky na zadanou pozici
// ji nekontroluje, zda pi tom nedochz k chyb - pedpokld, e kontrola probhla pedtm
void state_t::commit(int x, int y, fixed_piece_t& p, int num){
  for (int i=0; i<p.size; i++)
    for (int j=0; j<p.size; j++)
      if (p.a[i][j] && x+i<n  && y+j< m) b[x+i][y+j] = num;
}

// globln data pro prohledvn
free_piece_t pieces[MAXPIECES];
int pieces_count;
int solutions = 0, steps = 0;

// prohledvn metodou backtracking
void search(state_t& s, int num) {
  fixed_piece_t fp;
  steps++;
  
  if (num == pieces_count) {
    // nali jsme een 
    s.print();
    solutions++;
  } else {    
    // pro aktuln dlek vyzkoume vechny mon pozice v mce a
    // vechna mon natoen
    for (int i=0; i<s.n; i++)
      for (int j=0; j<s.m; j++)
	for (int p=0; p<pieces[num].positions; p++) {
	  pieces[num].get_rotation(p, fp);
	  if (s.match(i,j,fp)) {
	    state_t ss;
	    ss.copy(s);
	    ss.commit(i,j,fp,num+1);
	    search(ss, num+1);
	  }
	}
  }
}

int main(int argc, char** argv) {
  ifstream inpieces (argv[1]);
  ifstream inboard (argv[2]);

  if (!inpieces.is_open() || !inboard.is_open()) {
    cerr << "Program ocekava jako argument jmena dvou souboru (soubor s dilky a soubor s mrizkou)." <<endl;
    return 1;
  }

  inpieces >> pieces_count;
  for (int i=0; i < pieces_count; i++)
    pieces[i].read(inpieces);
  
  state_t init;
  init.read(inboard);
  
  search(init, 0);

  cout << "Pocet kroku pri prohledavani:\t"<<steps<<endl;
  cout << "Pocet nalezenych reseni:\t"<<solutions<<endl;
  return 0;
}
